/*
** Module   : p2j.canvas_renderer.js
** Abstract : GUI-specific canvas renderer
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 SBI 20150905 Created the canvas renderer in order to enclose all canvas drawing in one class.
*/

"use strict";

/**
 * Creates the canvas renderer.
 * 
 * @param    {HtmlCanvasElement} canvas
 *           The html canvas element.
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context on which to draw.
 * @param    {LinesStroke} strokesManager
 *           The stroke manager
 * @param    {Object} fontsManager
 *           The p2j fonts manager.
 * @param    {Object} logger
 *           The p2j logger.
 */
function CanvasRenderer(canvas, ctx, strokesManager, fontsManager, logger)
{
   this.canvas = canvas;
   this.ctx = ctx;
   this.strokesManager = strokesManager;
   this.logger = logger;
   this.fontsManager = fontsManager;

   /** Color scaling factor, compatible with Java2D. */
   var SCALE_FACTOR = 0.7;

   /** Minimum color value for darkening/lightening. */
   var MIN_COLOR_VALUE = Math.floor(1.0 / (1.0 - SCALE_FACTOR));

   /** Common data structure used in direct drawing operations. */
   this.pixel = this.ctx.createImageData(1, 1);
   
   /** Internal portion of the pixel data structure used in direct drawing operations. */
   this.pixData = this.pixel.data;
   
   // force alpha to 100% opaque
   this.pixData[3] = 0xFF;
   
   // eliminate anti-aliasing of images
   this.ctx.imageSmoothingEnabled = false;
   
   /** Tracks the current cumulative coordinate translation that is in effect. */
   this.origin = { "x" : 0, "y" : 0 };

   /** Stack of graphics context state, used for managing drawing, clipping and transforms. */
   this.state = new Array();

   /** Current color for direct drawing (in the form [ red, green, blue ]). */
   this.rawColor = new Array(0, 0, 0);

   /** The current window stroke style id. */
   this.strokeStyleId = strokesManager.getDefaultStrokeStyle();
   
   /** The current window stroke width. */
   this.strokeWidth = 1;

   //sets the default stroke for the canvas
   this.strokesManager.applyStrokeToCanvas(this.ctx, this.strokeStyleId, this.strokeWidth);

   // force all drawing to be inside pixel "cells" instead of "stradling" between two cells
   // which is the default; this eliminates much of the worst of the anti-aliasing and
   // positioning problems inherent in canvas operations; this will also save off the
   // current state of the graphics context so that it can be restored later 
   this.translate(0.5, 0.5);
   
   /**
    * Convert a standard fillStyle or strokeStyle color into an array of [ R, G, B ] values.
    *
    * @param    {String} spec
    *           Input string which must be in the form "#RRGGBB" where RR, GG and BB are 2-digit
    *           hexadecimal numbers.
    *
    * @return   The parsed raw color array of integers in [ R, G, B ] form.
    */
   this.parseColor = function(spec)
   {
      if (typeof spec !== "string")
      {
         p2j.logger.error("parseColor() spec must be a string");
         return null;                                 
      }
      
      if (spec.length !== 7)
      {
         p2j.logger.error("parseColor() spec must have length of 7");
         return null;
      }
      
      if (spec.charAt(0) !== "#")
      {
         p2j.logger.error("parseColor() spec must start with #");
         return null;
      }
      
      var r = parseInt(spec.substr(1, 2), 16);
      var g = parseInt(spec.substr(3, 2), 16);
      var b = parseInt(spec.substr(5, 2), 16);

      if (r === NaN || g === NaN || b === NaN)
      {
         return null;
      }
      
      return [ r, g, b ];
   };

   /**
    * Test if the given argument is an array of 3 integer values.
    *
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    *
    * @return   {Boolean} <code>true</code> if the argument is in [ R, G, B ] form.
    */
   this.validateColor = function(color)
   {
      if (!(color instanceof Array))
      {
         return false;
      }
      
      if (color.length !== 3)
      {
         return false;
      }
      
      if (typeof color[0] !== "number" ||
          typeof color[1] !== "number" ||
          typeof color[2] !== "number")
      {
         return false;
      }
      
      if (color[0] < 0 || color[0] > 255 ||
          color[1] < 0 || color[1] > 255 ||
          color[2] < 0 || color[2] > 255)
      {
         return false;
      }
      
      return true;
   };

   /**
    * Convert the given array of 3 integer values into an HTML color string.
    *
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    *
    * @return   {String} A valid HTML color string.
    */
   this.createColorString = function(color)
   {
      if (!this.validateColor(color))
      {
         return null;
      }
                                         
      return "rgb(" + color[0].toString() + ", "
                    + color[1].toString() + ", "
                    + color[2].toString() + ")";
   };

   /**
    * Convert the given array of RGB values into a lighter color in a manner that is compatible 
    * with Java2D.
    *
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    *
    * @return   {Number[]} A lighter shade of the same color.
    */
   this.lightenColor = function(color)
   {
      if (!this.validateColor(color))
      {
         return null;
      }
      
      // handle black directly, the result is NOT very much different but this is how it is
      // done in Java
      if (color[0] == 0 && color[1] == 0 && color[2] == 0)
         return [ MIN_COLOR_VALUE, MIN_COLOR_VALUE, MIN_COLOR_VALUE ];
      
      // force inputs to a minimim color value if between 0 and that minimum color value
      // (yes, this is how Java does it)
      var rIn = (color[0] > 0 && color[0] < MIN_COLOR_VALUE) ? MIN_COLOR_VALUE : color[0];
      var gIn = (color[1] > 0 && color[1] < MIN_COLOR_VALUE) ? MIN_COLOR_VALUE : color[1];
      var bIn = (color[2] > 0 && color[2] < MIN_COLOR_VALUE) ? MIN_COLOR_VALUE : color[2];
      
      // scale it
      var r = Math.min(Math.floor(rIn / SCALE_FACTOR), 255);
      var g = Math.min(Math.floor(gIn / SCALE_FACTOR), 255);
      var b = Math.min(Math.floor(bIn / SCALE_FACTOR), 255);
      
      return [ r, g, b ];
   };

   /**
    * Convert the given array of RGB values into a darker color in a manner that is compatible 
    * with Java2D.
    *
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    *
    * @return   {Number[]} A darker shade of the same color.
    */
   this.darkenColor = function(color)
   {
      if (!this.validateColor(color))
      {
         return null;
      }
      
      var r = Math.max(0, Math.floor(color[0] * SCALE_FACTOR));
      var g = Math.max(0, Math.floor(color[1] * SCALE_FACTOR));
      var b = Math.max(0, Math.floor(color[2] * SCALE_FACTOR));
      
      return [ r, g, b ];
   };
}

/**
 * Fill or stroke the current path on the given context, first using the current path as a
 * clipping region.  This ensures that the drawing of the path will not extend outside of the
 * path by a half-pixel (which it would otherwise). The state of the context is saved on
 * entry and restored after the drawing operation, so there will be no residual clipping
 * region when this function returns.
 *
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context on which to draw.
 * @param    {Boolean} fill
 *           <code>true</code> to fill the drawn rectangle with the given color.
 */
CanvasRenderer.prototype.renderClosedPath = function(ctx, fill)
{
   ctx.save();
   ctx.clip();
   
   if (fill)
   {
      ctx.fill();
   }
   else
   {
      ctx.stroke();
   }
   
   ctx.restore();
};

/**
 * Adjust the origin for pixel drawing on the canvas.
 *
 * @param    {Number} x
 *           The translate along X-axis.
 * @param    {Number} y
 *           The translate along Y-axis.
 */
CanvasRenderer.prototype.adjustOrigin = function(x, y)
{
   this.origin.x += x;
   this.origin.y += y;
};

/**
 * Directly draw the specified pixel in the given color.
 *
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context on which to draw.
 * @param    {Number} x
 *           X coordinate of the pixel to be drawn.
 * @param    {Number} y
 *           Y coordinate of the pixel to be drawn.
 * @param    {Number[]} color
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 */
CanvasRenderer.prototype.drawPixel = function(ctx, x, y, color)
{
   this.pixData[0] = color[0];
   this.pixData[1] = color[1];
   this.pixData[2] = color[2];
   
   this.ctx.putImageData(this.pixel, x, y);
};

/**
 * Draw a line in the given color from (x1, y1) to (x2, y2) inclusive, using direct drawing.
 * <p>
 * This is an implementation of the well known 50+ year old Bresenham algorithm.
 *
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context on which to draw.
 * @param    {Number} x1
 *           X coordinate of the starting pixel to be drawn.
 * @param    {Number} y1
 *           Y coordinate of the starting pixel to be drawn.
 * @param    {Number} x2
 *           X coordinate of the ending pixel to be drawn.
 * @param    {Number} y2
 *           Y coordinate of the ending pixel to be drawn.
 * @param    {Number[]} color
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 * 
 * @return   {Array} path
 *           The {x:.,y:.} point per a pixel array that represents the drawing line segment.
 */
CanvasRenderer.prototype.drawLineSegment = function(ctx, x1, y1, x2, y2, color)
{
   x1 = x1 + this.origin.x;
   x2 = x2 + this.origin.x;
   
   y1 = y1 + this.origin.y;
   y2 = y2 + this.origin.y;

   // holds the line pixels to draw
   var path = [];
   var dx = Math.abs(x2 - x1);
   var dy = Math.abs(y2 - y1);
   
   var xIncr = x1 < x2 ? 1 : -1;
   var yIncr = y1 < y2 ? 1 : -1;
   
   dy = -dy;
   
   var err = dx + dy;
   var err2;
   
   var x = x1;
   var y = y1;
   
   var tooMany = 0;
   
   while (true)
   {
      if (this.strokesManager.isDirectDrawingStrokeStyle(this.strokeStyleId))
      {
         this.drawPixel(ctx, x, y, color);
      }
      path.push({x : x, y : y});
      
      tooMany++;
      
      if (x == x2 && y == y2)
      {
         break;
      }
      else
      {
         if (tooMany > 10000)
         {
            this.logger.logFormatted("Bresenham is out of control for x1 = %d, y1 = %d,"
                     + " x2 = %d, y2 = %d and current x = %d, y = %d!",[x1, y1, x2, y2, x, y]);
            console.trace();
            break;
         }
      }
         
      err2 = err * 2;
      
      if (err2 >= dy)
      {
         err += dy;
         x   += xIncr; 
      }
      
      if (err2 <= dx)
      {
         err += dx;
         y   += yIncr;
      }
   }
   
   return path;
};

/**
 * Draw a line in the given color from (x1, y1) to (x2, y2) inclusive, using direct drawing and
 * add its pixels path to the end of the provided pixels path.
 * <p>
 * This is an implementation of the well known 50+ year old Bresenham algorithm.
 *
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context on which to draw.
 * @param    {Number} x1
 *           X coordinate of the starting pixel to be drawn.
 * @param    {Number} y1
 *           Y coordinate of the starting pixel to be drawn.
 * @param    {Number} x2
 *           X coordinate of the ending pixel to be drawn.
 * @param    {Number} y2
 *           Y coordinate of the ending pixel to be drawn.
 * @param    {Number[]} color
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 * @param    {Array} path
 *           The provided holder of {x:.,y:.} points, the target drawing as a set of sequential
 *           line points will be added at the end of the provided path.
 * @return   {Array}
 *           The {x:.,y:.} point per a pixel array that represents the drawing line segment.
 */
CanvasRenderer.prototype.drawLine = function(ctx, x1, y1, x2, y2, color, path)
{
   var segment = this.drawLineSegment(ctx, x1, y1, x2, y2, color);
   Array.prototype.push.apply(path, segment);
   return segment;
}

/**
 * Stroke a line in the given color from (x1, y1) to (x2, y2) inclusive, using direct drawing and
 * add its pixels path to the end of the provided pixels path.
 * <p>
 * This is an implementation of the well known 50+ year old Bresenham algorithm.
 *
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context on which to draw.
 * @param    {Number} x1
 *           X coordinate of the starting pixel to be drawn.
 * @param    {Number} y1
 *           Y coordinate of the starting pixel to be drawn.
 * @param    {Number} x2
 *           X coordinate of the ending pixel to be drawn.
 * @param    {Number} y2
 *           Y coordinate of the ending pixel to be drawn.
 * @param    {Number[]} color
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 * @param    {Array} path
 *           The provided holder of {x:.,y:.} points, the target drawing as a set of sequential
 *           line points will be added at the end of the provided path.
 */
CanvasRenderer.prototype.strokeLineSegment = function(ctx, x1, y1, x2, y2, color)
{
   var path = this.drawLineSegment(ctx, x1, y1, x2, y2, color);
   this.strokesManager.applyStrokeToPath(this.ctx,
         this.strokeStyleId,
         this.strokeWidth,
         color,
         path);
}

/**
 * Draw a rectangle in the given color and dimensions with the line drawing being overwritten
 * using direct drawing to eliminate the negative/unwanted effects of alni-aliasing.
 *
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context on which to draw.
 * @param    {Number} x
 *           X coordinate of the top left corner.
 * @param    {Number} y
 *           Y coordinate of the top left corner.
 * @param    {Number} width
 *           Width of the rectangle in pixels.
 * @param    {Number} height
 *           Height of the rectangle in pixels.
 * @param    {Number[]} color
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 *           This will be used for the border lines and when in fill mode, also the interior.
 * @param    {Boolean} fill
 *           <code>true</code> to fill the drawn rectangle with the given color.
 * 
 * @return   {Array} path
 *           The {x:.,y:.} point per a pixel array that represents the rectangle outerline.
 */
CanvasRenderer.prototype.drawRect = function(ctx, x, y, width, height, color, fill)
{
   // filled rectangles draw 1 pixel smaller in 2 dimensions than stroked rectangles
   var inset = 0;
   var path  = [];
   // use vector operations for the interior of the rectangle
   if (fill)
   {
      inset = 1;
      ctx.beginPath();
      ctx.rect(x, y, width - inset, height - inset);
      this.renderClosedPath(ctx, fill);
   }
   
   // now overdraw the stroked portion to eliminate anti-aliasing, we draw in a
   // clockwise direction (since we are not using paths, this is not strictly necessary)
   this.drawLine(ctx, x, y, x + width - inset, y, color, path);
   this.drawLine(ctx, x + width - inset, y + 1, x + width - inset, y + height - inset, color, path);
   this.drawLine(ctx, x + width - 1 - inset, y + height - inset, x + 1, y + height - inset, color, path);
   this.drawLine(ctx, x, y + height - inset, x, y + 1, color, path);// close the path
   
   return path;
};

/**
 * Stroke a rectangle in the given color and dimensions with the line drawing being overwritten
 * using direct drawing to eliminate the negative/unwanted effects of alni-aliasing.
 *
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context on which to draw.
 * @param    {Number} x
 *           X coordinate of the top left corner.
 * @param    {Number} y
 *           Y coordinate of the top left corner.
 * @param    {Number} width
 *           Width of the rectangle in pixels.
 * @param    {Number} height
 *           Height of the rectangle in pixels.
 * @param    {Number[]} color
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 *           This will be used for the border lines and when in fill mode, also the interior.
 * @param    {Boolean} fill
 *           <code>true</code> to fill the drawn rectangle with the given color.
 */
CanvasRenderer.prototype.strokeRect = function(ctx, x, y, width, height, color, fill)
{
   // filled rectangles draw 1 pixel smaller in 2 dimensions than stroked rectangles
   var inset = 0;
   var path  = [];
   // use vector operations for the interior of the rectangle
   if (fill)
   {
      inset = 1;
      ctx.beginPath();
      ctx.rect(x, y, width - inset, height - inset);
      this.renderClosedPath(ctx, fill);
   }
   
   // now overdraw the stroked portion to eliminate anti-aliasing, we draw in a
   // clockwise direction (since we are not using paths, this is not strictly necessary)
   this.drawLine(ctx, x, y, x + width - inset, y, color, path);
   this.drawLine(ctx, x + width - inset, y + 1, x + width - inset, y + height - inset, color, path);
   this.drawLine(ctx, x + width - 1 - inset, y + height - inset, x + 1, y + height - inset, color, path);
   this.drawLine(ctx, x, y + height - inset, x, y + 1, color, path);// close the path
   
   this.strokesManager.applyStrokeToPath(this.ctx,
         this.strokeStyleId,
         this.strokeWidth,
         color,
         path);
};

/**
 * It draws an image on the canvas.
 * 
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context on which to draw.
 * @param    {Number} x
 *           X coordinate of the top left corner.
 * @param    {Number} y
 *           Y coordinate of the top left corner.
 * @param    {Number} width
 *           Width of the image in pixels.
 * @param    {Number} height
 *           Height of the image in pixels.
 * @param    {Array} imgData
 *           The image data array contains the encoded image started from the given offset.
 * @param    {Number} imgDataOffset
 *           The offset of the encoded image data.
 */
CanvasRenderer.prototype.drawImage = function(ctx, x, y, width, height, imgData, imgDataOffset)
{
   /**
    * Pixels must be returned as non-premultiplied alpha values according to:
    * http://www.w3.org/TR/2dcontext/#pixel-manipulation
    */
   var img = ctx.getImageData(this.origin.x + x, this.origin.y + y, width, height);
   var data = img.data;
   var pixelsInBytes = width * height * 4;
   var offset = imgDataOffset;
   for (var p = 0; p < pixelsInBytes; p += 4, offset += 4)
   {
      var alpha = imgData[offset + 3];
      var beta = 256 - alpha;
      if (alpha == 255)
      {
         /** server sends non-premultiplied image data */
         data[p]     = imgData[offset];
         data[p + 1] = imgData[offset + 1];
         data[p + 2] = imgData[offset + 2];
         data[p + 3] = imgData[offset + 3];
      }
      else if (alpha > 0)
      {
         /** server sends non-premultiplied image data */
         data[p]     = (imgData[offset]     * alpha + data[p]     * beta) >> 8;
         data[p + 1] = (imgData[offset + 1] * alpha + data[p + 1] * beta) >> 8;
         data[p + 2] = (imgData[offset + 2] * alpha + data[p + 2] * beta) >> 8;
         data[p + 3] = (imgData[offset + 3] * alpha + data[p + 3] * beta) >> 8;
      };
   };
   ctx.putImageData(img, this.origin.x + x, this.origin.y + y);
};

/**
 * Draw a rounded rectangle with the given context, dimensions and fill.
 *
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context.
 * @param    {Number} x
 *           X coordinate of the top left of the rectangle (if a square corner was placed
 *           there).
 * @param    {Number} y
 *           Y coordinate of the top left of the rectangle (if a square corner was placed
 *           there).
 * @param    {Number} width
 *           Width of the rectangle.
 * @param    {Number} height
 *           Height of the rectangle.
 * @param    {Number} diameter
 *           Diameter of the arc that defines the rounded corner.
 * @param    {Number[]} color  
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 * @param    {Boolean} fill
 *           <code>true</code> to fill the rectangle with the current color, otherwise just
 *           stroke the rectangle.
 */
CanvasRenderer.prototype.drawRoundRect = function(ctx, x, y, width, height, diameter, color, fill)
{
   var radius = diameter / 2;
   
   // arc control points
   var northWestX = x;
   var northWestY = y;
   var northEastX = x + width;
   var northEastY = y;
   var southWestX = x;
   var southWestY = y + height;
   var southEastX = x + width;
   var southEastY = y + height;
   
   // line end points                                       
   var topLeftX     = northWestX + radius;
   var topLeftY     = northWestY;
   var topRightX    = northEastX - radius;
   var topRightY    = northEastY;
   var rightUpX     = northEastX;
   var rightUpY     = northEastY + radius;
   var rightDownX   = southEastX;
   var rightDownY   = southEastY - radius;
   var bottomRightX = southEastX - radius;
   var bottomRightY = southEastY;
   var bottomLeftX  = southWestX + radius;
   var bottomLeftY  = southWestY;
   var leftDownX    = southWestX;
   var leftDownY    = southWestY - radius;
   var leftUpX      = northWestX;
   var leftUpY      = northWestY + radius;
   
   ctx.beginPath();
   ctx.moveTo(topRightX, topRightY);
   ctx.quadraticCurveTo(northEastX, northEastY, rightUpX, rightUpY);
   ctx.lineTo(rightDownX, rightDownY);
   ctx.quadraticCurveTo(southEastX, southEastY, bottomRightX, bottomRightY);
   ctx.lineTo(bottomLeftX, bottomLeftY);
   ctx.quadraticCurveTo(southWestX, southWestY, leftDownX, leftDownY);
   ctx.lineTo(leftUpX, leftUpY);
   ctx.quadraticCurveTo(northWestX, northWestY, topLeftX, topLeftY);
   ctx.closePath();
   
   this.renderClosedPath(ctx, fill);
   
   // overdraw the anti-aliased line segments              
   var path = [];
   this.drawLine(ctx, topLeftX, topLeftY, topRightX, topRightY, color, path);
   this.drawLine(ctx, rightUpX, rightUpY, rightDownX, rightDownY, color, path);
   this.drawLine(ctx, bottomLeftX, bottomLeftY, bottomRightX, bottomRightY, color, path);
   this.drawLine(ctx, leftUpX, leftUpY, leftDownX, leftDownY, color, path);
   this.strokesManager.applyStrokeToPath(this.ctx,
         this.strokeStyleId,
         this.strokeWidth,
         color,
         path);
};

/**
 * Draw a slightly shaded "3D" rectangle with the given context, dimensions and fill. In
 * raised mode the rectangle is drawn with a lighter color used for the top and left borders
 * and any fill that is needed, while the bottom and right borders are a darker version of
 * the given color.  If not raised, a darker shade of the given color is used for the top
 * and left borders and any requested fill, while a lighter shade is used for the bottom and
 * right borders.
 *
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context.
 * @param    {Number} x
 *           X coordinate of the top left of the rectangle.
 * @param    {Number} y
 *           Y coordinate of the top left of the rectangle.
 * @param    {Number} width
 *           Width of the rectangle in pixels.
 * @param    {Number} height
 *           Height of the rectangle in pixels.
 * @param    {Number[]} color  
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 * @param    {Boolean} fill
 *           <code>true</code> to fill the rectangle with the current color, otherwise just
 *           stroke the rectangle.
 * @param    {Boolean} raise
 *           <code>true</code> to draw in raised mode.
 */
CanvasRenderer.prototype.draw3DRect = function(ctx, x, y, width, height, color, fill, raised)
{
   // save off our colors
   var oriFillColor   = ctx.fillStyle;
   var oriStrokeColor = ctx.strokeStyle;
   
   var darker  = this.darkenColor(color);
   var lighter = this.lightenColor(color);
   
   // save off our state so that we can clear our clipping region on exit
   ctx.save();
   
   // the Java implementation draws 1 pixel larger in each dimension
   width  = width + 1;
   height = height + 1;
   
   // setup our clipping region, this is needed because the line width (1 pixel wide) is
   // drawn half on one side of hte path and half on the other side; this makes paths a
   // half pixel bigger than the caller would expect; by defining the same path as a
   // clipping region, the half pixel width that is outside the path is never output
   ctx.beginPath();
   ctx.rect(x, y, width, height);
   ctx.clip();
   
   var inset = 0;
   
   var current = raised ? color : darker;
   
   // fill the rectangle with the original color (raised mode) or the darker color otherwise
   if (fill)
   {
      inset = 1;
      ctx.fillStyle = this.createColorString(current);
      ctx.beginPath();
      ctx.rect(x, y, width - 2, height - 2);
      this.renderClosedPath(ctx, fill);
   }
   
   current = raised ? lighter : darker;
   
   // draw the top and left sides in the contrasting color (raised mode) or the darker color
   // otherwise
   var path = [];// create new pixels path to draw
   this.drawLine(ctx, x, y, x + width - 2 - inset, y, current, path);
   this.drawLine(ctx, x, y + 1, x, y + height - 1 - inset, current, path);
   this.strokesManager.applyStrokeToPath(this.ctx,
         this.strokeStyleId,
         this.strokeWidth,
         current,
         path);

   // draw the bottom and right sides in the darker color (raised mode) or the contrasting
   // color otherwise
   current = raised ? darker : lighter;
   path = []; // create new pixels path to draw
   this.drawLine(ctx, x + 1, y + height - 1 - inset, x + width - 1 - inset, y + height - 1 - inset, current, path);
   this.drawLine(ctx, x + width - 1 - inset, y, x + width - 1 - inset, y + height - 2, current, path);
   this.strokesManager.applyStrokeToPath(this.ctx,
         this.strokeStyleId,
         this.strokeWidth,
         current,
         path);
   
   // clear the clipping region
   ctx.restore();
   
   // restore our colors
   ctx.fillStyle   = oriFillColor;
   ctx.strokeStyle = oriStrokeColor;
};

/**
 * Draw a closed polygon with the given context, vertices and fill.
 *
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context.
 * @param    {Number} xPoints
 *           X coordinates of the points.
 * @param    {Number} yPoints
 *           Y coordinates of the points.
 * @param    {Number} num
 *           Number of points to draw.
 * @param    {Number[]} color  
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 * @param    {Boolean} fill
 *           <code>true</code> to fill the rectangle with the current color, otherwise just
 *           stroke the rectangle.
 * @return   {Array}
 *           The {x:.,y:.} point per a pixel array that represents this poligon.
 */
CanvasRenderer.prototype.drawPolygon = function(ctx, xPoints, yPoints, num, color, fill)
{
   var i;

   // use vector operations for the interior of the poygon
   if (fill)
   {
      ctx.beginPath();
      ctx.moveTo(xPoints[0], yPoints[0]);
      for (i = 1; i < num; i++)
      {
         ctx.lineTo(xPoints[i], yPoints[i]);
      }
      ctx.closePath();
      this.renderClosedPath(ctx, fill);
   }
   var path = [];
   // now overdraw the stroked portion
   for (i = 0; i < (num - 1); i++)
   {
      this.drawLine(ctx, xPoints[i], yPoints[i], xPoints[i + 1], yPoints[i + 1], color, path);
   }
   return path;
};

/**
 * Stroke a closed polygon with the given context, vertices and fill.
 *
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context.
 * @param    {Number} xPoints
 *           X coordinates of the points.
 * @param    {Number} yPoints
 *           Y coordinates of the points.
 * @param    {Number} num
 *           Number of points to draw.
 * @param    {Number[]} color  
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 * @param    {Boolean} fill
 *           <code>true</code> to fill the rectangle with the current color, otherwise just
 *           stroke the rectangle.
 */
CanvasRenderer.prototype.strokePolygon = function(ctx, xPoints, yPoints, num, color, fill)
{
   var path = this.drawPolygon(ctx, xPoints, yPoints, num, color, fill);
   this.strokesManager.applyStrokeToPath(this.ctx,
         this.strokeStyleId,
         this.strokeWidth,
         color,
         path);
};

/**
 * Set the color for future graphics operations on this window.
 *
 * @param    {Number[]} or {String} color
 *           This must be either a valid array of 3 integers between 0 and 255 inclusive OR it
 *           an HTML color specification in "#RRGGBB" form.
 */
CanvasRenderer.prototype.setColor = function(color)
{
   var colorSpec;

   if (color instanceof Array)
   {
      if (this.validateColor(color))
      {
         colorSpec = this.createColorString(color);
         
         // store the color in a form that can be used for direct drawing
         this.rawColor = [ color[0], color[1], color[2] ];
      }
      else
      {
         this.logger.error("Invalid array form of color specification.");
         return;
      }
   }
   else
   {
      if (typeof color === "string")                  
      {
         colorSpec = color;
         
         // store the color in a form that can be used for direct drawing
         this.rawColor = this.parseColor(color);
      }
      else
      {
         this.logger.error("Invalid string form of color specification.");
         return;
      }
   }
   
   // change the canvas colors
   this.ctx.fillStyle   = colorSpec;
   this.ctx.strokeStyle = colorSpec;
};

/**
 * Sets the stroke style.
 *                              
 * @param    {Number} strokeStyleId
 *           The predefined stroke style.
 */
CanvasRenderer.prototype.setStrokeStyleId = function(strokeStyleId)
{
   this.strokeStyleId = strokeStyleId;
};

/**
 * Sets the stroke line width.
 *                              
 * @param    {Number} width
 *           The stroke line width.
 */
CanvasRenderer.prototype.setStrokeWidth = function(width)
{
   this.strokeWidth = width;    
};


/**
 * Sets the line stroke.
 * 
 * @param    {Number} strokeStyleId
 *           The predefined stroke style id.
 * @param    {Number} width
 *           The stroke line width.
 */
CanvasRenderer.prototype.setLineStroke = function(strokeStyleId, width)
{                               
   this.setStrokeStyleId(strokeStyleId);
   this.setStrokeWidth(width);
   this.strokesManager.applyStrokeToCanvas(this.ctx, strokeStyleId, width);
};

/**
 * Add the given rectangle to the current clipping region.
 * 
 * @param    {Number} x
 *           The top left X coordinate in pixels.
 * @param    {Number} y
 *           The top left Y coordinate in pixels.
 * @param    {Number} width
 *           The rectangle width in pixels.
 * @param    {Number} height
 *           The rectangle height in pixels.
 */
CanvasRenderer.prototype.clip = function(x, y, width, height)
{
   this.save({ clip : { "x" : x, "y" : y, "width" : width, "height" : height } });
   this.ctx.beginPath();
   this.ctx.rect(x, y, width, height);
   this.ctx.clip();
};

/**
 * Remove the most recently added rectangle from the current clipping region. The graphics
 * context state is popped to restore the previously used values.
 */
CanvasRenderer.prototype.unclip = function()
{
   this.restore();
};

/**
 * Save graphics context state to the stack.
 *
 * @param    {Object} additional
 *           Caller provided name/value pairs to add to the state.
 */
CanvasRenderer.prototype.save = function(additional)
{
   var cur = 
   {
      fillstyle     : this.ctx.fillStyle,
      strokeStyle   : this.ctx.strokeStyle,
      strokeStyleId : this.strokeStyleId,
      strokeWidth   : this.strokeWidth
   };
   
   if (additional)
   {
      for (var prop in additional)
      {
         if (additional.hasOwnProperty(prop))
         {
            cur[prop] = additional[prop];
         }
      }
   }
   
   this.state.push(cur);
   
   // now force the context to push its state internally (only the context itself can save off
   // its accumulated clip region and its cumulative transforms)
   this.ctx.save();
};

/**
 * Restore the graphics context state to the state as stored on the top of the stack.
 * <p>
 * The state in our own custom state stack is not actually used for the restore, instead the
 * context's internal stack is used (ctx.restore()).  Our stack is used for other purposes
 * and we simply clear it here.
 */
CanvasRenderer.prototype.restore = function()
{
   this.state.pop();
   this.ctx.restore();
};

/**
 * Obtain the graphics context state as stored on the top of the stack.
 *
 * @return   The top of state stack object.
 */
CanvasRenderer.prototype.peek = function()
{
   var last = this.state.length - 1;
   
   return last >= 0 ? this.state[last] : {};
};

/**
 * Replay the saved graphics context state back into a newly reset graphics context. This
 * should only be used after a canvas resize operation.
 */
CanvasRenderer.prototype.replay = function()
{
   var len = this.state.length;
   
   for (var i = 0; i < len; i++)
   {
      var next = this.state[i];
      
      this.ctx.fillstyle   = next.fillStyle;
      this.ctx.strokeStyle = next.strokeStyle;
      this.setLineStroke(next.strokeStyleId, next.strokeWidth);
      var x;
      var y;
      var width;
      var height;
      
      if (next.translate)
      {
         x = next.translate.x;
         y = next.translate.y;
         
         // don't call the this.translate() method here because it modifies the stack that
         // we are simply trying to replay; we don't want the stack to be modified
         this.ctx.translate(x, y);
      }
   
      if (next.clip)
      {
         x      = next.clip.x;
         y      = next.clip.y;
         width  = next.clip.width;
         height = next.clip.height;                               
                                                                  
         // don't call the this.translate() method here because it modifies the stack that
         // we are simply trying to replay; we don't want the stack to be modified
         this.ctx.beginPath();
         this.ctx.rect(x, y, width, height);
         this.ctx.clip();
      }
   }
};

/**
 * Untranslate the graphics context by the given X and Y factors. This does a restore on
 * the previously saved context state.  It will error out if some sanity checks don't pass.
 * Make sure that calls to this function are always properly paired with equal and opposite
 * calls to translate().
 *
 * @param    {Number} x
 *           X dimension factor.
 * @param    {Number} y
 *           Y dimension factor.
 */
CanvasRenderer.prototype.untranslate = function(x, y)
{
   var top = this.peek();
   
   if (top.translate)
   {
      if (top.translate.x === -x && top.translate.y === -y)
      {
         // clear our state back to the version that doesn't have this translate
         this.restore();
         
         this.adjustOrigin(x, y);
      }
      else
      {
         this.logger.error("Mismatching translate pop values, expected (" +
                  top.translate.x.toString() + ", " + top.translate.y.toString() +
                  ") and found " + x.toString() + ", " + y.toString() + ").");
      }
   }
   else
   {
      this.logger.error("Out of order translate pop!");
   }
};

/**
 * Resize the canvas to the given width and height.
 *
 * @param    {Number} width
 *           The new canvas width in pixels.
 * @param    {Number} height
 *           The new canvas height in pixels.
 */
CanvasRenderer.prototype.resize = function(width, height)
{
   // save off the previously drawn content
   var oldPixels = this.ctx.getImageData(0, 0, this.canvas.width, this.canvas.height);
   
   // this is expensive, but we really do need to change the overall size of the canvas 
   this.canvas.width  = width;
   this.canvas.height = height;
   
   // all of the current state including colors, line stroke, translation and clipping
   // must be reapplied to the context because changing the size resets everything
   this.replay();
                                
   // re-blit the previously drawn content
   this.ctx.putImageData(oldPixels, 0, 0);
};   

/**
 * Translate the graphics context by the given X and Y factors.  This will also save the state
 * of the context before the translate is applied, so that the translate values can be
 * reversed later.
 *
 * @param    {Number} x
 *           X dimension factor.
 * @param    {Number} y
 *           Y dimension factor.
 */
CanvasRenderer.prototype.translate = function(x, y)
{
   this.save({ translate : { "x" : x, "y" : y } });
   
   this.adjustOrigin(x, y);
   
   this.ctx.translate(x, y);
};

/**
 * Draws a given text at the given (x,y) position.
 * 
 * @param    {Number} text
 *           The text to draw.
 * @param    {Number} x
 *           The X-axis of the coordinate for the text starting point.
 * @param    {Number} y
 *           The Y-axis of the coordinate for the text starting point or the height
 *           of the text rectangle from the most top screen position.
 * @param    {Boolean} centered
 *           Defines the text base line and the interpretation of y.
 *           If it is true, the text base line is set to 'middle' and interprets y as a height,
 *           otherwise the text base line is set to 'bottom'.
 */
CanvasRenderer.prototype.drawText = function(text, x, y, centered)
{
   if (centered)
   {
      // in this case, the "y" is the height in which it needs to be vertically centered
      y = y / 2;
      
      this.ctx.textBaseline = 'middle';
   }
   else
   {
      this.ctx.textBaseline = 'bottom';
   }
   
   this.ctx.fillText(text, x, y);
}

/**
 * Draws a given text at the given (x,y) position fitted to the given text rectangle.
 * 
 * @param    {Number} currentFont
 *           The current font ID, in the js font-table.
 * @param    {Number} text
 *           The text to draw.
 * @param    {Number} x
 *           The X-axis of the coordinate for the text starting point.
 * @param    {Number} y
 *           The Y-axis of the coordinate for the text starting point or the height
 *           of the text rectangle from the most top screen position.
 * @param    {Boolean} centered
 *           Defines the text base line and the interpretation of y.
 *           If it is true, the text base line is set to 'middle' and interprets y as a height,
 *           otherwise the text base line is set to 'bottom'.
 * @param    {Number} lwidth
 *           The space width to be used to display the given text.
 *           The text is scaled to fit this width, squeezed or extended
 * @param    {Number} lheight
 *           The space height to be used to display the given text.
 *           The text is scaled to fit this height, squeezed or extended.
 */
CanvasRenderer.prototype.drawScaledText = function(currentFont, text, x, y, centered, lwidth, lheight)
{
   var textWidth = this.fontsManager.getTextWidth(currentFont, text);
   var textHeight = this.fontsManager.getTextHeight(currentFont, text);
   
   var widthScale = lwidth / textWidth;
   var heightScale = lheight / textHeight;
   
   // scale drawing context to computed width
   // this scales the desired structure.x as well so it needs adjusting
   this.ctx.save();
   this.ctx.scale(widthScale, heightScale);
   
   var scaleBackWidth = 1 / widthScale;
   var scaleBackHeight = 1 / heightScale;
   
   if (centered)
   {
      // in this case, the "y" is the height in which it needs to be vertically centered
      y = y / 2;

      this.ctx.textBaseline = 'middle';
   }
   else
   {
      this.ctx.textBaseline = 'bottom';
   }

   this.ctx.fillText(text, x * scaleBackWidth, y * scaleBackHeight);
   this.ctx.restore();
}