/*
** Module   : p2j.fonts.js
** Abstract : Font and text metrics helpers.
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 CA  20150812 First version.
** 002 CA  20150818 Fixed param/return data types in the function javadoc.
**                  Added font maxWidth.
** 003 CA  20150826 If a font witht the same specifications already exists, re-use it from the 
**                  local font-table by sending its index to the caller.
**                  Fixed font scaling (use window.devicePixelRatio).
*/

"use strict";

p2j.fonts = (function()
{
   /** Font object which will be returned from this initialization function. */
   var me = {}; 

   /* 
    * A js-side font-table of known fonts.  Provides a 1-to-1 mappings with the fonts known
    * known by the java side.
    *
    * This data structure contains:
    * 1. keys: a numeric value starting from 0  
    * 2. each value referenced by the key contains the font details, via an associative array:
    *    - "name"
    *    - "size"
    *    - "bold"
    *    - "italic"
    *    - "underline"
    *    - "font" - the full font name, to be used by the context
    *    - "height": font height
    *    - "widths": font widths
    *    - "maxWidth": font max width
    */
   var fontTable = {};

   /** The canvas used as a worker for measuring texts. */
   var fontCanvas;

   /** The canvas context. */
   var fontCtx;

   /**
    * Initialize module.
    */
   me.init = function()
   {
      fontCanvas = document.createElement('canvas');
      fontCtx = fontCanvas.getContext('2d');
   };

   /**
    * Create a font with the given specs and registered it in the font table.
    * 
    * @param    {Number} font
    *           The font identifier.
    * @param    {String} name
    *           The font name.
    * @param    {Number} size
    *           The font size, in points.
    * @param    {Number} style
    *           The serialized font style.
    * @param    {String} b64Font
    *           The base64-encoded font definition.  Empty string if not available.
    */
   me.createFont = function(font, name, size, style, b64Font)
   {
      var bold      = ((style & (1 << 0)) == (1 << 0));
      
      var italic    = ((style & (1 << 1)) == (1 << 1));
      var underline = ((style & (1 << 2)) == (1 << 2));
      
      var theFont = (bold ? "bold " : "") +
                    (italic ? "italic " : "") +
                    /* (underline ? "underline " : "") + */ 
                    size + "pt" +
                    " '" + name + "'";
      p2j.logger.log(theFont);
      // check if there is already a font with the same details
      var fontKey = theFont.toLowerCase();
      var fcandidate = fontTable[fontKey];
      if (fcandidate != undefined)
      {
         return fcandidate;
      }
      
      var fheight = textHeight(theFont, '\u2503Eghy', 128);
      var fwidths = fontWidths(theFont);
      var maxWidth = 0;
      
      for (var i = 0; i < fwidths.length; i++)
      {
         if (fwidths[i] > maxWidth)
         {
            maxWidth = fwidths[i];
         }
      }
      
      var fontDef = 
      {
         "name"      : name,
         "size"      : size,
         "bold"      : bold,
         "italic"    : italic,
         "underline" : underline,
         "font"      : theFont,
         "height"    : fheight,
         "widths"    : fwidths,
         "maxWidth"  : maxWidth
      };
      
      fontTable[font] = fontDef;
      fontTable[fontKey] = font;
      
      if (b64Font != "")
      {
         var cssFont = "@font-face {  font-family: '" + name + "';";
         cssFont = cssFont + " src: url('data:application/x-font-truetype;base64," + b64Font + "')";
         cssFont = cssFont + " format('truetype'); }";

         var newStyle = document.createElement('style');
         newStyle.appendChild(document.createTextNode(cssFont));
         document.head.insertBefore(newStyle, document.head.childNodes[0]);
      }
      
      return font;
   };
   
   /**
    * Derive the font with the given ID to have the new size.
    * 
    * @param    {Number} font
    *           The font identifier.
    */
   me.deriveFont = function(font)
   {
      var fd = fontTable[font];
      var size = Math.round(window.devicePixelRatio * fd.size);
      
      if (fd.size != size)
      {
         fd.size = size;

         var theFont = (fd.bold ? "bold " : "") +
                       (fd.italic ? "italic " : "") +
                       /* (fd.underline ? "underline " : "") + */ 
                       fd.size + "pt" +
                       " '" + fd.name + "'";
         
         // set the key for the new name
         fontTable[theFont.toLowerCase()] = font;
         
         fd.font = theFont;
      }
   };
   
   /**
    * For the specified font in the font table, set its style.
    * 
    * @param    {Number} font
    *           The font identifier.
    * @param    {Number} style
    *           The serialized font style.
    * 
    * @return   {boolean} true if the font style has changed.
    */
   me.setFontStyle = function(font, style)
   {
      var bold      = ((style & (1 << 0)) == (1 << 0));
      var italic    = ((style & (1 << 1)) == (1 << 1));
      var underline = ((style & (1 << 2)) == (1 << 2));

      var fd = fontTable[font];
      var changed = false;
      
      if (fd.bold != bold)
      {
         fd.bold = bold;
         changed = true;
      }
      
      if (fd.italic != italic)
      {
         fd.italic = italic;
         changed = true;
      }
      
      if (fd.underline != underline)
      {
         fd.underline = underline;
         changed = true;
      }
      
      if (changed)
      {
         var theFont = (fd.bold ? "bold " : "") +
                       (fd.italic ? "italic " : "") +
                       /* (fd.underline ? "underline " : "") + */ 
                       fd.size + "pt" +
                       " '" + fd.name + "'";
   
         fd["font"] = theFont;
      }
      
      return changed;
   };
   
   /**
    * Get the name of the font associated with the font ID.
    * 
    * @param    {Number} font
    *           The font identifier.
    *
    * @return   {String} The font name.
    */
   me.getFontName = function(font)
   {
      return fontTable[font].font;
   };
   
   /**
    * Get the widths of the font associated with the font ID.
    * 
    * @param    {Number} font
    *           The font identifier.
    *
    * @return   {Number[]} The font widths.
    */
   me.getFontWidths = function(font)
   {
      return fontTable[font].widths;
   };
   
   /**
    * Get the height of the font associated with the font ID.
    * 
    * @param    {Number} font
    *           The font identifier.
    *
    * @return   {Number} The font height.
    */
   me.getFontHeight = function(font)
   {
      return fontTable[font].height;
   };
   
   /**
    * Get the text width using the font associated with the font ID.
    * 
    * @param    {Number} font
    *           The font identifier.
    * @param    {String} text
    *           The text.
    *
    * @return   {Number} The text width.
    */
   me.getTextWidth = function(font, text)
   {
      if (text.length == 0)
      {
         return 0;
      }
   
      var f = fontTable[font];

      return stringWidth(f.font, text);
   };
   
   /**
    * Get the text height using the font associated with the font ID.
    * 
    * @param    {Number} font
    *           The font identifier.
    * @param    {String} text
    *           The text.
    *
    * @return   {Number} The text height.
    */
   me.getTextHeight = function(font, text)
   {
      if (text.length == 0)
      {
         return 0;
      }
   
      var f = fontTable[font];

      text = '\u2503' + text;

      // ensure the text will fit the canvas
      var pxWidth = f.maxWidth * text.length;
      
      var theight = textHeight(f.font, text, pxWidth);
      
      return theight;
   };
 
   /**
    * Compute the height of the text, using the given font.
    *
    * @param    {Number} font
    *           The font, as in the fontTable[x].font.
    * @param    {String} text
    *           The text.
    * @param    {Number} width
    *           The canvas width, so that the text will fit.
    *
    * @return   {Number} The text height.
    */
   function textHeight(font, text, width)
   {
      var height = 128;

      fontCtx.save();

      fontCanvas.setAttribute('height', height);
      fontCanvas.setAttribute('width', width);
      fontCtx.fillStyle = 'black';
      fontCtx.fillRect(0, 0, width, height);
      fontCtx.textBaseline = 'top';
      fontCtx.fillStyle = 'white';
      fontCtx.font = font;
      fontCtx.fillText(text, 0, 0);

      var pixels = fontCtx.getImageData(0, 0, width, height).data;

      // row numbers where we first find letter end where it ends 
      var beg = -1;
      var end = -1;

      for (var row = 0; row < height; row++)
      {
         for (var col = 0; col < width; col++)
         {
            var index = (row * width + col) * 4;
            
            // if pixel is not white (background color)
            if (pixels[index] === 0)
            {
               // we haven't met white (font color) pixel
               // on the row and the letters was detected
               if (col === (width - 1) && beg !== -1)
               {
                  end = row;
                  row = height;
                  break;
               }
            }
            else
            {
               // we find top of letter
               if (beg === -1)
               {
                  beg = row;
               }
               // ..letters body
               break;
            }
         }
      }

      fontCtx.restore();
      
      return end - beg;
   }
   
   /**
    * Compute the font widths, for the given font.
    *
    * @param    {Number} font
    *           The font, as in the fontTable[x].font.
    * 
    * @return   {Number[]} The font widths.
    */
   function fontWidths(font)
   {
      var widths = [];
      for (var ch = 0 ; ch < 256 ; ch++)
      {
          widths.push(stringWidth(font, String.fromCharCode(ch)));
      }
      return widths;
   }
   
   /**
    * Measure the width of the text, using the given font.
    *
    * @param    {Number} font
    *           The font, as in the fontTable[x].font.
    * @param    {String} text
    *           The text.
    * 
    * @return   {Number} The text width.
    */
   function stringWidth(font, text)
   {
      fontCtx.save();

      fontCtx.textBaseline = 'top';
      fontCtx.font = font;
      
      var width = fontCtx.measureText(text).width;
      
      fontCtx.restore();

      return parseInt(width);
   }

   return me;
}());
