/*
** Module   : p2j.logger.js
** Abstract : Application logger.
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 SBI  20150905 The first version of the p2j logger.
*/

"use strict";

p2j.logger = (function()
{
   /** Generic logger. */
   var me = {};
   
   /** Storage for log entries. */
   var logstack = new Array();
   
   /** The stack capacity, the last visible history.*/
   var maxStackSize;
   /**
    * Initialize the logger with the storage capacity.
    * 
    * @param    {Number}
    *           The log stack capacity.
    */
   me.init = function(capacity)
   {
      maxStackSize = capacity;
   };

   /**
    * Adds messages to the log stack.
    * 
    * @param    {String} msg
    *           The text message to log.
    */
   me.log = function(msg)
   {
      save(msg);
   };
   
   /*
    * Log formatted message, supported format is %d.
    * 
    * @param    {String} format
    *           The formatted message with %d place holders.
    * @param    {Array} parameters
    *           The parameters are listed according their positions in the formatted message.
    */
   me.logFormatted = function(format, parameters)
   {
      save(format.split("%d").map(
            function(currValue, index, arr)
            {
               if (this[index] === undefined)
               {
                  return currValue;
               }
               return currValue + this[index];
            }, parameters).reduce(
                  function(prevValue, currValue)
                  {
                     return prevValue + currValue;
                  }, ""));
   };
   
   /**
    * Log an error message.
    * 
    * @param    {string} text
    *           The error text to display.
    */
   me.error = function(text)
   {
      save(text);
      console.error(text);
      if (p2j.sound)
      {
         p2j.sound.beep();
      }
   };
   
   /**
    * Saves messages to the log stack.
    * 
    * @param    {String} msg
    *           The text message to log.
    */
   function save(msg)
   {
      var size = logstack.push(msg);
      
      /**
       * if the stack size exceeds 2 times the maximal stack size, then it is shifted to the left
       * and its size is reduced to the maximal stack size.
       */
      if (size > 2 * maxStackSize)
      {
         logstack.splice(0, maxStackSize);
      }
   };
   
   /**
    * Display a prompt that shows the current log output and allows it to be copied to the
    * system clipboard.
    */
   me.viewLog = function()
   {
      var text = "<!DOCTYPE html>\n"                         +
                 "<html>\n"     +
                 "   <head>\n"                               +
                 "      <title>Log Output</title>\n"         +
                 "   </head>\n"                              +
                 "   <body>\n"                               +
                 "      <textarea rows=\"10\" cols=\"60\">\n";

      for (var i = 0; i < logstack.length; i++)
      {
         text = text + "\n" + logstack[i];
      }
      
      text = text                  +
             "      </textarea>\n" +
             "   </body>\n"        +
             "</html>\n";
      
      window.open("data:text/html;charset=utf-8," + encodeURIComponent(text),
                  "",
                  "width=500,height=500");
   };
   return me;
}());
